// pg io simulator.
//
// we simulate an index scan, sort of.
// given multiple files it more acts like one (fetching from index, fetching
// from the heap).
//
// we also support random writing. 
//
// the general idea is to open a number of files (hopefully larger han ram)
// and randomly seek around and read blocks. in write mode a percentage of
// the blocks are rewritten.
// 
// released under the BSD license.
// comments/questions: Jeff Trout - jeff@jefftrout.com
//
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <signal.h>

#define TV2DOUBLE(tv) ((double)(tv)->tv_sec + (double)((tv)->tv_usec) / 1000000)
int stop = 0;

typedef struct _FileInfo
{
  int fd;
  int numBlocks;
  int curBlock;
} FileInfo;


// doSIGINT
//
// called on ctrl-c to stop
//
void doSIGINT(int x)
{
  printf("Interrupt\n");
  stop = 1;
}


// getFile
//
// open a file, return its FileInfo structure
//
FileInfo *getFile(char *fn, int writeMode)
{
  FileInfo *f;
  struct stat sinfo;
  int flags = O_RDONLY;

  if(writeMode) 
	{
	  flags = O_RDWR;
	}

  f = malloc(sizeof(FileInfo));
  stat(fn, &sinfo);

  f->numBlocks = sinfo.st_size / 8192;
  f->fd = open(fn, flags);
  f->curBlock = 0;
  if(f->fd <= 0)
	{
	  printf("open failed: %s\n", fn);
	  exit(0);
	}
  
  return f;
}

int main(int argc, char **argv)
{
  int count;
  char buf[8193];
  FileInfo *files[25]; // up to 25 files we can use. 
  int numFiles;
  int x;
  off_t cb;
  int n;
  int seqScan = 0;
  int numReads = 0;
  int writeMode = 0;
  int writePercent = 0;
  int writeBlocks = 0;
  char ch;
  FileInfo *f;
  struct timeval tv;
  double start;
  double end;
  double elapsed;

  signal(SIGINT, doSIGINT);

  while((ch = getopt(argc, argv, "b:sw:")) != EOF)
	{
	  switch(ch)
		{
		case 'b':
		  numReads = atoi(optarg);
		  break;
		case 's':
		  printf("Seq Scan\n");
		  seqScan = 1;
		  break;
		case 'w':
		  writeMode = 1;
		  if(optarg)
			{
			  writePercent = atoi(optarg);
			}
		  printf("Write Mode: %d%%\n", writePercent);
		  break;
		  
		}
	}

  argv += optind;
  argc -= optind;

  printf("Arg: %d\n", argc);
  for(numFiles = 0; numFiles < argc; numFiles++)
	{
	  printf("Added %s\n", argv[numFiles]);
	  files[numFiles] = getFile(argv[numFiles], writeMode);
	}

  srand(getpid());

  gettimeofday(&tv, NULL);
  start = TV2DOUBLE(&tv);

  // randomly read the file(s).
  for(x = 0; x < numReads && !stop; x++)
	{
	  f = files[rand() % numFiles];

	  // pick a file
	  if(seqScan)
		{
		  cb = f->curBlock % f->numBlocks;
		  f->curBlock++;
		}
	  else
		{
		  cb = rand() % f->numBlocks;
		}
	  
	  // this shouldn't happen
	  if(cb > f->numBlocks) 
		{
		  printf("Dork\n");
		}

	  if(lseek(f->fd, (off_t)8192 * cb, SEEK_SET) == -1)
		{
		  printf("Seek %d failed: %d %s\n", x, cb * 8192, strerror(errno));
		}

	  n = read(f->fd, buf, 8192);

	  //	  printf("N: %d\n", n);
	  if(n != 8192)
		{
		  printf("Bad read: %s\n", strerror(errno));
		}

	  if(writeMode)
		{
		  // we just write the block back.
		  if(rand() % 100 < writePercent)
			{
			  if(lseek(f->fd, (off_t)8192 * cb, SEEK_SET) == -1)
				{
				  printf("Seek %d failed: %d %s\n", x, cb * 8192, strerror(errno));
				}
			  if(write(f->fd, buf, 8192) != 8192)
				{
				  printf("Write Failed: %s\n", strerror(errno));
				  exit(0);
				}
			  writeBlocks++;
			}
		}
	}

  numReads = x;

  for(x = 0; x < numFiles; x++)
	{
	  close(files[x]->fd);
	}

  gettimeofday(&tv, NULL);
  end = TV2DOUBLE(&tv);

  elapsed = end - start;
  printf("Elapsed: %0.2f\n", elapsed);
  printf("Read %d blocks Wrote %d blocks\n", numReads, writeBlocks);
  printf("%0.2f op/sec, %0.2fkB/sec\n", (double)numReads / elapsed, 
		 (((double)numReads * 8192) / elapsed) / 1024);

}

